unit CoaThds;

// Copyright  2001 by Ziff Davis Media, Inc.
// Written by Neil J. Rubenking

interface
uses Windows, SysUtils, Classes, StdCtrls, CoaGenl;
type
  // The AddType function is used in a Synchronize() call to send
  // an item back to the main program
  AddType = procedure (const cItem, cPath, cSub1, cSub2, cOld,
    cNew : String; vItmType, vData1, vHits, vCount, vStatus : Integer;
    vData2 : DWORD) OF Object;
  // The FailBackType function is used in a Synchronize() call to
  // tell the main program to set the last I items to failure status
  FailBackType    = procedure(I, stat : Integer) OF Object;
  // The ReportCountType function is used in a Synchronize() call to
  // tell the main program another 100 items have been processed
  ReportCountType = procedure(I : Integer) OF Object;

  TCOAThread = class(TThread)
  protected
    fFindLong     : String; // find this long filename and...
    fReplLong     : String; // ... replace with this long filename
    fFindShrt     : String; // find this short filename and...
    fReplShrt     : String; // ... replace with this short filename
    fSlash        : Boolean; // true if both adddresses end in slash
    fReportCount  : ReportCountType; // function to report progress
    fAddItem      : AddType;         // function to report a hit
    fFailBack     : FailBackType;    // function to fail items ex post facto
    // values passed to fAddItem
    fCurPath      : String;  // pathname or root key
    fCurItem      : String;  // filename or registry keypath
    fSub1         : String;  // INI section or registry key
    fSub2         : String;  // INI key or registry value
    fOld          : String;  // string containing old address
    fNew          : String;  // string with new address swapped in
    fCount        : Integer; // number of files/keys tested
    fHits         : Integer; // number of "hits"
    fHitStart     : Integer; // start of hit-set, for failback
    fStatus       : Integer; // see statXxxx constants
    fChgType      : Integer; // just what was changed?
    fData1        : Integer; // INI line-number, LNK icon index,
                             // Registry root-key value
    fData2        : DWORD;   // INI section line, Registry data type
    procedure SyncAddItem;
    procedure SyncCount;
    procedure SyncFailBack;
    function DoFind(const S : String; VAR short, slash : Boolean) :
      Boolean; virtual;
  public
    constructor Create(const vFindL, vFindS, vReplL, vReplS : String;
      vAddItem : AddType; vReportCount : ReportCountType;
      vFailBack : FailBackType);
  end;

  TCOAFileThread = class(TCOAThread)
  protected
    fExt : String;
    // Each descendant of TCOAFileThread must implement SearchFile()
    procedure SearchFile(const S : String; Atts : DWORD); virtual;
      abstract;
    procedure SearchPath(const Pth : String);
    procedure SearchPaths(const Pth : String);
    procedure SearchAll;
  public
    constructor Create(const vFindL, vFindS, vReplL, vReplS,
      vExt : String; vAddItem : AddType;
      vReportCount : ReportCountType; vFailBack : FailBackType);
    procedure Execute; override;
  END;

  TChangeObj = class(TObject)
  // This object is used to store detailed information about a
  // changed item. It is attached to the Data field of listview
  // items, and it is streamed to/from change record files
  protected
    fTyp     : Integer;
    fLoc     : String;  // pathname or root key
    fItm     : String;  // filename or registry keypath
    fSub1    : String;  // INI section or registry key
    fSub2    : String;  // INI key or registry value
    fOld     : String;  // string containing old address
    fNew     : String;  // string with new address swapped in
    fData1    : Integer; // INI line-number, LNK icon index,
                        // Registry root-key value
    fData2 : DWORD;     // INI section line, Registry data type
    fUndo    : Integer; // undo-status

    function GetHint: String; virtual; abstract;
  public
    constructor Create(vTyp, vData1, vData2, vStatus : Integer;
      const vLoc, vItm, vSub1, vSub2, vOld, vNew : String);
    constructor CreateForStream;
    procedure SaveToStream(Stream : TStream);
    procedure LoadFromStream(Stream : TStream);
    procedure SetFail(newStat : Integer);
    // Every descendant of TChangeObj must implement Undo and
    // the two reporting methods
    function Undo : Boolean; virtual; abstract;
    procedure ReportText(VAR T : TextFile; VAR MainS : String);
      virtual; abstract;
    procedure ReportHTML(VAR T : TextFile; VAR MainS : String);
      virtual; abstract;

    property Hint : String read GetHint;
    property   OldS : String read fOld;
    property   NewS : String read fNew;
    property   LocS : String read fLoc;
    property   ItmS : String read fItm;
    property   Sub1 : String read fSub1;
    property UndoStatus : Integer read fUndo;
  end;

implementation
uses coafuncs;

constructor TCOAThread.Create(const vFindL, vFindS, vReplL,
  vReplS : String; vAddItem : AddType;
  vReportCount : ReportCountType; vFailBack : FailBackType);
begin
  Inherited Create(True);
  FreeOnTerminate := True;
  fFindLong       := UpperCase(vFindL);
  fFindShrt       := UpperCase(vFindS);
  fSlash          := HasSlash(vFindL) AND HasSlash(vFindS);
  fReplLong       := vReplL;
  fReplShrt       := vReplS;
  fAddItem        := vAddItem;
  fReportCount    := vReportCount;
  fFailBack       := vFailBack;
  fCount          := 0;
  fHits           := 0;
  fStatus         := 0;
  fData2          := 0;
  fSub1           := '';
  fSub2           := '';
end;

procedure TCOAThread.SyncAddItem;
begin
  Inc(fHits);
  fAddItem(fCurItem, fCurPath, fSub1, fSub2, fOld, fNew, fChgType,
    fData1, fHits, fCount, fStatus, fData2);
end;

procedure TCOAThread.SyncCount;
begin
  fReportCount(fCount);
end;

function TCOAThread.DoFind(const S : String; VAR short,
  slash : Boolean) : Boolean;
begin
  short  := False;
  slash  := False;
  Result := False;
  // IF the long name is found in S, we have a hit
  IF Pos(fFindLong, UpperCase(S)) > 0 THEN
    Result := True
  // IF the short name is found in S, we have a short hit
  ELSE IF (fFindShrt <> '') AND (Pos(fFindShrt, UpperCase(S)) > 0) THEN
    begin
      Result := True;
      short  := True;
    end
  // IF the name ends in a slash, try another thing
  ELSE IF fSlash THEN
    begin
      // If S exactly equals the long name before the slash, a hit
      IF (Length(S) + 1 = Length(fFindLong)) AND
        (Pos(UpperCase(S), fFindLong) = 1) THEN
        begin
          slash  := True;
          Result := True;
        end
      // If S exactly equals the short name before the slash, a hit
      ELSE IF (fFindShrt <> '') AND
        (Length(S) + 1 = Length(fFindShrt)) AND
        (Pos(UpperCase(S), fFindShrt) = 1) THEN
        begin
          slash  := True;
          short  := True;
          Result := True;
        end;
    end;
end;

(**** TCOAFileThread ****)
constructor TCOAFileThread.Create(const vFindL, vFindS, vReplL,
  vReplS, vExt : String; vAddItem : AddType;
  vReportCount : ReportCountType; vFailBack : FailBackType);
begin
  Inherited Create(vFindL, vFindS, vReplL, vReplS, vAddItem,
    vReportCount, vFailBack);
  fExt  := vExt;
end;

procedure TCOAFileThread.SearchPath(const Pth : String);
// Find all files with the specified extension in the supplied path
// and pass them to the SearchFile() method. The SearchFile()
// method modifies files, and this can cause the modified file to
// get put back into the FindFirstFile / FindNextFile queue. To
// avoid repeated changes, this function first puts all the matching
// files into a list and then processes them.
VAR
  FH : THandle;
  FD : TWin32FindData;
  TS : TStringList;
  N  : Integer;
begin
  fCurPath := Pth;
  fCurItem := '';
  TS := TStringList.Create;
  try
    FH := FindFirstFile(PChar(Pth+'\*.'+fExt), FD);
    IF FH = INVALID_HANDLE_VALUE THEN Exit;
    try
      REPEAT
        IF FD.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY > 0 THEN
          Continue;
        TS.AddObject(FD.cFilename, Pointer(FD.dwFileAttributes));
      UNTIL (NOT FindNextFile(FH, FD));
    finally
      Windows.FindClose(FH);
    end;
    FOR N := 0 TO TS.Count-1 DO
      begin
        Inc(fCount);
        IF fCount MOD 100 = 0 THEN
          Synchronize(SyncCount);
        SearchFile(TS[N], DWORD(TS.Objects[N]));
      end;
  finally
    TS.Free;
  end;
end;

procedure TCOAFileThread.SearchPaths(const Pth : String);
// Search all folders contained in the supplied path
const
  cHidSys = FILE_ATTRIBUTE_SYSTEM OR FILE_ATTRIBUTE_HIDDEN;
VAR
  FH : THandle;
  FD : TWin32FindData;
begin
  SearchPath(Pth);
  FH := FindFirstFile(PChar(Pth+'\*.*'), FD);
  IF FH = INVALID_HANDLE_VALUE THEN Exit;
  try
    REPEAT
      IF FD.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY > 0 THEN
        begin
          // Skip the "self" and "parent" entries
          IF StrComp(FD.cFilename, '.') = 0 THEN Continue;
          IF StrComp(FD.cFilename, '..') = 0 THEN Continue;
          // Skip the recycle bin
          try
            IF (Length(Pth) = 2) AND
              ((StrIComp(FD.cFilename, 'RECYCLED') = 0) OR
               (StrIComp(FD.cFilename, 'RECYCLER') = 0)) AND
              (FD.dwFileAttributes AND cHidSys = cHidSys) THEN
              Continue;
          except
            ON E:Exception DO
              Continue;
          end;
          SearchPaths(Pth + '\' + FD.cFilename);
        end;
    UNTIL (NOT FindNextFile(FH, FD));
  finally
    Windows.FindClose(FH);
  end;
end;

procedure TCOAFileThread.SearchAll;
VAR
  LogicalDrives : ARRAY[0..128] OF Char;
  Drv           : PChar;
begin
  GetLogicalDriveStrings(128, LogicalDrives);
  Drv := LogicalDrives;
  WHILE (Drv[0] <> #0) AND (NOT Terminated) DO
    begin
      IF GetDriveType(Drv) = DRIVE_FIXED THEN
        SearchPaths(Copy(StrPas(Drv), 1, 2));
      Drv := StrEnd(Drv)+1;
    end;
  Synchronize(SyncCount);
end;

procedure TCOAFileThread.Execute;
begin
  SearchAll;
end;

{ TChangeObj }

constructor TChangeObj.Create(vTyp, vData1, vData2, vStatus: Integer;
  const vLoc, vItm, vSub1, vSub2, vOld, vNew: String);
begin
  Inherited Create;
  fTyp   := vTyp;
  fData1 := vData1;
  fData2 := vData2;
  fLoc   := vLoc;
  fItm   := vItm;
  fSub1  := vSub1;
  fSub2  := vSub2;
  fOld   := vOld;
  fNew   := vNew;
  fUndo  := vStatus;
end;

constructor TChangeObj.CreateForStream;
// Just create the object - its data will be initialized by
// LoadFromStream()
begin
  inherited Create;
end;

procedure TChangeObj.LoadFromStream(Stream: TStream);
VAR Reader : TReader;
begin
  Reader := TReader.Create(Stream, 4096);
  WITH Reader DO
  try
    fTyp   := ReadInteger;
    fLoc   := ReadString;
    fItm   := ReadString;
    fSub1  := ReadString;
    fSub2  := ReadString;
    fOld   := ReadString;
    fNew   := ReadString;
    fData1 := ReadInteger;
    fUndo  := ReadInteger;
    fData2 := DWORD(ReadInt64);
  finally
    Free;
  end;
end;

procedure TChangeObj.SaveToStream(Stream: TStream);
VAR Writer : TWriter;
begin
  Writer := TWriter.Create(Stream, 4096);
  WITH Writer DO
  try
    WriteInteger(fTyp);
    WriteString(fLoc);
    WriteString(fItm);
    WriteString(fSub1);
    WriteString(fSub2);
    WriteString(fOld);
    WriteString(fNew);
    WriteInteger(fData1);
    WriteInteger(fUndo);
    WriteInteger(fData2);
  finally
    Free;
  end;
end;

procedure TChangeObj.SetFail(newStat : Integer);
begin
  fUndo := newStat;
end;

procedure TCOAThread.SyncFailBack;
begin
  IF fHits > fHitStart THEN
    fFailBack(fHits - fHitStart, fStatus);
end;

end.
