unit Thd_Reg;

// Copyright  2001 by Ziff Davis Media, Inc.
// Written by Neil J. Rubenking

interface
uses
  Windows, SysUtils, Classes, CoaThds;

type
  TSearchRegThread = class(TCOAThread)
  protected
    fIgnores : TStringList;
    fFindLongU     : String; // Unicode long filename
    fFindShrtU     : String; // Unicode short filename
  public
    constructor Create(const vFindL, vFindS, vReplL,
      vReplS : String;
      vAddItem : AddType; vReportCount : ReportCountType; vIgnores : TStrings;
      vFailBack : FailBackType);
    destructor Destroy; override;
    procedure Execute; override;
  END;

  TChangeRegObj = class(TChangeObj)
  protected
    function GetHint : String; override;
  public
    function Undo :  Boolean; override;
    procedure ReportText(VAR T : TextFile;
      VAR MainS : String); override;
    procedure ReportHTML(VAR T : TextFile;
      VAR MainS : String); override;
  end;

implementation
Uses Registry, CoaGenl, coafuncs;

function StatString(stat : Integer) : String;
begin
  CASE stat AND $F OF
    statBlank : Result := '';
    statCheck : Result := '';
    statUndoX : begin
      Result := 'Status: Undo attempt failed';
      CASE stat AND (NOT $F) OF
        statFaildData : Result := Result +
          '; could not write Registry data';
        statFaildWrit : Result := Result +
          '; could not open Registry key for writing';
        statFaildSame : Result := Result +
          '; new value name already exists';
        statFaildName : Result := Result +
          '; could not rename Registry value';
      END;
    end;
    statUndOK : Result := 'Status: Undo successful';
    statFaild : begin
      Result := 'Status: Change attempt failed';
      CASE stat AND (NOT $F) OF
        statFaildData : Result := Result +
          '; could not write Registry data';
        statFaildWrit : Result := Result +
          '; could not open Registry key for writing';
        statFaildSame : Result := Result +
          '; new value name already exists';
        statFaildName : Result := Result +
          '; could not rename Registry value';
      END;
    end;
    statNSafe : begin
      Result := 'Status: Change not attempted - unsafe';
      CASE stat AND (NOT $F) OF
        statNsafeBina : Result := Result + ' to change binary data';
        statNSafeRKey : Result := Result + ' to change key name';
      END;
    end;
  END;
end;

type
  THoldObj = class(TObject)
  public
    fCode  : Integer;
    fData1 : String;
    fType  : DWORD;
    constructor Create(vCode : Integer; vType : DWORD; vData : String);
  end;

constructor TSearchRegThread.Create(const vFindL, vFindS,
  vReplL, vReplS :
  String; vAddItem : AddType; vReportCount : ReportCountType; vIgnores : TStrings;
      vFailBack : FailBackType);
VAR N : Integer;
begin
  Inherited Create(vFindL, vFindS, vReplL, vReplS, vAddItem, vReportCount, vFailBack);
  fIgnores := TStringList.Create;
  fIgnores.Assign(vIgnores);
  SetLength(fFindLongU, Length(fFindLong)*2);
  FOR N := 1 TO Length(fFindLong) DO
    begin
      fFindLongU[N*2]   := #0;
      fFindLongU[N*2-1] := fFindLong[N];
    end;
  SetLength(fFindShrtU, Length(fFindShrt)*2);
  FOR N := 1 TO Length(fFindShrt) DO
    begin
      fFindShrtU[N*2]   := #0;
      fFindShrtU[N*2-1] := fFindShrt[N];
    end;
end;

destructor TSearchRegThread.Destroy;
begin
  fIgnores.free;
  inherited;
end;

procedure TSearchRegThread.Execute;
VAR
  R : TRegistry;
  RootK : String;

  procedure SearchKey(const Pth, vKeyStr : String; level : Integer);
  const
    foundName = 1;
    shortName = 2;
    foundData = 4;
    shortData = 8;
    foundBina = 16;
    shortBina = 32;
    slashAny  = 64;
  VAR
    S        : String;
    N        : Integer;
    TL       : TStringList;
    short    : Boolean;
    slash    : Boolean;

    function DoFindUni(const S : String; VAR short, slash : Boolean) :
      Boolean;
    // Build a simple-minded Unicode string by interpolating ASCII
    // #0 chars between each char of the criterion string, and
    // check whether this string is present
    begin
      short  := False;
      slash  := False;
      Result := False;
      // IF the findlong is found in S, we have a hit
      IF Pos(fFindLongU, UpperCase(S)) > 0 THEN
        Result := True
      // IF the findshrt is found in S, we have a short hit
      ELSE IF (fFindShrtU <> '') AND (Pos(fFindShrtU, UpperCase(S)) > 0) THEN
        begin
          Result := True;
          short  := True;
        end
      // IF the findlong ends in a slash, try another thing
      ELSE IF fSlash THEN
        begin
          // If S exactly equals the findlong before the slash, good
          IF (Length(S) + 1 = Length(fFindLongU)) AND
            (Pos(UpperCase(S), fFindLongU) = 1) THEN
            begin
              slash := True;
              Result := True;
            end
          // If S exactly equals the findshrt before the slash, good
          ELSE IF (fFindShrtU <> '') AND (Length(S) + 1 = Length(fFindShrtU)) AND
            (Pos(UpperCase(S), fFindShrtU) = 1) THEN
            begin
              slash := True;
              short := True;
              Result := True;
            end;
        end;
    end;

    function DoFindBin(vBuff : PByteArray; vSize : Integer;
      VAR short, slash : Boolean) : Boolean;
    VAR
      DatS : String;
    begin
      SetLength(datS, vSize);
      Move(vBuff^, datS[1], vSize);
      IF DoFind(DatS, short, slash) THEN
        Result := True
      ELSE
        Result := DoFindUni(DatS, short, slash);
    end;

    function FindEither(Idx : Integer) : Boolean;
    // Search the value found at index Idx for the old address.
    // Also search text data. In the case of the data, if not
    // found try searching as unicode, and searching binary,
    // but don't change Unicode or binary. If anything is found
    // add a THoldObj to the list item containing code for
    // details.
    VAR
      short : Boolean;
      slash : Boolean;
      code  : Integer;
      DType : DWORD;
      DSize : DWORD;
      DBuff : PByteArray;
      DatS  : String;
    begin
      code  := 0;
      DatS  := '';
      DType := 0;
      IF DoFind(TL[Idx], short, slash) THEN
        begin
          code := code OR foundName;
          IF short THEN code := code OR shortName;
          IF slash THEN code := code OR slashAny;
        end;
      IF RegQueryValueEx(R.CurrentKey, PChar(TL[Idx]), nil,
        @DType, nil, @Dsize) = ERROR_SUCCESS THEN
        begin
          GetMem(DBuff, DSize);
          try
            IF RegQueryValueEx(R.CurrentKey, PChar(TL[Idx]), nil,
              @DType, @DBuff[0], @Dsize) = ERROR_SUCCESS THEN
              begin
                CASE DType OF
                  REG_SZ, REG_EXPAND_SZ : begin
                    DatS := StrPas(PChar(DBuff));
                    IF DoFind(DatS, short, slash) THEN
                      begin
                        code := code OR foundData;
                        IF short THEN code := code OR shortData;
                        IF slash THEN code := code OR slashAny;
                      end
                    ELSE IF DoFindUni(DatS, short, slash) THEN
                      begin
                        code := code OR foundBina;
                        IF short THEN code := code OR shortBina;
                        IF slash THEN code := code OR slashAny;
                      end
                  end;
                  ELSE begin
                    IF DoFindBin(DBuff, DSize, short, slash) THEN
                      begin
                        DatS := '(binary data)';
                        code := code OR foundBina;
                        IF short THEN code := code OR shortBina;
                        //ignore slash
                      end
                  end;
                END;
              end;
          finally
            FreeMem(DBuff);
          end;
        end;
      IF Code <> 0 THEN
        TL.Objects[Idx] := THoldObj.Create(code, DType, DatS);
      Result := Code <> 0;
    end;

    procedure MakeChanges(idx : Integer; able : Boolean);
    VAR
      THO   : THoldObj;
      DType : DWORD;
      DSize : DWORD;
      Rslt  : DWORD;
      DBuff : PByteArray;
    begin
      THO   := THoldObj(TL.Objects[idx]);
      fSub2 := TL[Idx];
      IF THO.fcode AND foundData > 0 THEN
        begin
          fChgType := chgRegData;
          fOld := THO.fData1;
          fData2 := THO.fType;
          fNew := StrChangeSmart(fOld, fFindLong, fReplLong,
            fFindShrt, fReplShrt, THO.fCode AND shortData > 0,
            THO.fCode AND slashAny > 0);
          IF able THEN
            begin
              DType := THO.fType;
              DSize := Length(fNew)+1;
              GetMem(DBuff, DSize);
              try
                StrPCopy(PChar(DBuff), fNew);
                Rslt := RegSetValueEx(R.CurrentKey, PChar(fSub2), 0,
                  DType, @DBuff[0], DSize);
                IF Rslt = ERROR_SUCCESS THEN
                  fStatus := statBlank
                ELSE
                  fStatus := statFaild OR statFaildData;
              finally
                FreeMem(DBuff);
              end;
            end
          ELSE fStatus := statFaild OR statFaildWrit;
          Synchronize(SyncAddItem);
        end;
      IF THO.fcode AND foundBina > 0 THEN
        begin
          fChgType := chgRegData;
          fOld     := THO.fData1;
          fData2   := THO.fType;
          IF THO.fCode AND shortData > 0 THEN
            fNew   := '(short name found)'
          ELSE
            fNew   := '(long name found)';
          fStatus  := statNSafe OR statNSafeBina;
          Synchronize(SyncAddItem);
        end;
      IF THO.fcode AND foundName > 0 THEN
        begin
          fChgType  := chgRegValue;
          fOld      := fSub2;
          fNew      := StrChangeSmart(fOld, fFindLong, fReplLong,
            fFindShrt, fReplShrt, THO.fCode AND shortName > 0,
            THO.fCode AND slashAny > 0);
          IF NOT able THEN
            fStatus := statFaild OR statFaildWrit
          ELSE IF R.ValueExists(fNew) THEN
            fStatus := statFaild OR statFaildSame
          ELSE
            begin
              fStatus := statFaild OR statFaildName;
              // To change the name, create a new value with the
              // same data and then delete the old
              IF RegQueryValueEx(R.CurrentKey, PChar(fold), nil,
                @DType, nil, @Dsize) = ERROR_SUCCESS THEN
              begin
                GetMem(DBuff, DSize);
                try
                  IF RegQueryValueEx(R.CurrentKey, PChar(fOld), nil,
                    @DType, @DBuff[0], @Dsize) = ERROR_SUCCESS THEN
                  begin
                    Rslt := RegSetValueEx(R.CurrentKey, PChar(fNew),
                      0, DType, @DBuff[0], Dsize);
                    IF Rslt = ERROR_SUCCESS THEN
                      begin
                        Rslt := RegDeleteValue(R.CurrentKey,
                          PChar(fOld));
                        IF Rslt = ERROR_SUCCESS THEN
                          fStatus := statBlank;
                      end;
                  end;
                finally
                  FreeMem(DBuff);
                end;
              end;
            end;
          Synchronize(SyncAddItem);
        end;
      THO.Free;
      TL.Objects[Idx] := nil;
    end;

  begin
    Inc(fCount);
    IF fCount MOD 100 = 0 THEN
      Synchronize(SyncCount);
    fCurItem := Pth;
    fSub1    := vKeyStr;
    S        := RootK + Pth;
    // If this key is on the equivalents list, ignore it
    FOR N := 0 TO fignores.Count-1 DO
      IF AnsiCompareText(S, fignores[N]) = 0 THEN
        Exit;
    fSub2 := '';
    IF DoFind(S, short, slash) THEN
      // Found in KEY - report, but fail it
      begin
        fChgType := chgRegKey;
        fOld     := S;
        fNew     := StrChangeSmart(fOld, fFindLong, fReplLong,
          fFindShrt, fReplShrt, short, slash);
        fStatus  := statNSafe OR statNSafeRKey;
        Synchronize(SyncAddItem);
      end;
    // If we can't open the key, skip it. Note that it gets opened
    // for read-only unless we actually need to write
    IF NOT R.OpenKeyReadOnly(Pth) THEN
      Exit;
    TL := TStringList.Create;
    try
      R.GetValueNames(TL);
      TL.Sort;
      // Test whether the old address is found anywhere in each
      // value or its data. If not, discard the value from the
      // list
      FOR N := TL.Count-1 DOWNTO 0 DO
        IF NOT FindEither(N) THEN
          TL.Delete(N);
      // IF there are any left, try to re-open the key for
      // read-write access and make the changes
      IF TL.Count > 0 THEN
        begin
          R.CloseKey;
          R.Access := KEY_ALL_ACCESS;
          IF R.OpenKey(Pth, False) THEN
            begin
              FOR N := 0 TO TL.Count-1 DO
                MakeChanges(N, True);
              R.CloseKey;
            end
          ELSE
            FOR N := 0 TO TL.Count-1 DO
              MakeChanges(N, False);
          R.OpenKeyReadOnly(Pth);
        end;
      FOR N := TL.Count-1 DOWNTO 0 DO
        THoldObj(TL.Objects[N]).Free;
      TL.Clear;
      // Now do the subkeys
      R.GetKeyNames(TL);
      TL.Sort;
      FOR N := 0 TO TL.Count-1 DO
        begin
          IF Terminated THEN Exit;
          IF TL[N] = '' THEN
            Continue;
          SearchKey(Pth+'\'+TL[N], TL[N], Level+1);
        end;
    finally
      TL.Free;
    end;
  end;

  procedure DoRoot(RK : HKey; const rkName : String);
  begin
    fData1    := RK;
    fCurPath  := rkName;
    RootK     := rkName;
    R.RootKey := RK;
    SearchKey('', '', 0);
  end;

begin
  R := TRegistry.Create;
  try
    DoRoot(HKEY_CLASSES_ROOT,     'HKEY_CLASSES_ROOT');
    IF NOT Terminated THEN
      DoRoot(HKEY_CURRENT_USER,   'HKEY_CURRENT_USER');
    IF NOT Terminated THEN
      DoRoot(HKEY_CURRENT_CONFIG, 'HKEY_CURRENT_CONFIG');
    IF NOT Terminated THEN
      DoRoot(HKEY_LOCAL_MACHINE,  'HKEY_LOCAL_MACHINE');
    IF NOT Terminated THEN
      DoRoot(HKEY_USERS,          'HKEY_USERS');
  finally
    R.Free;
    Synchronize(SyncCount);
  end;
end;

{ TChangeRegObj }

function TChangeRegObj.GetHint: String;
begin
  Result := fLoc+fItm+#13#10+'  *  *  *'+#13#10;
  CASE fTyp OF
    chgRegKey   : Result := Result + 'Changed: Key name';
    chgRegValue : Result := Result +
      Format('Changed: Value name in key "%s', [fSub1]);
    chgRegData  : Result := Result +
      Format('Changed: Data for value "%s" in key "%s"',
        [fSub2, fSub1]);
  END;
  CASE fUndo AND $F OF
    statBlank : ;
    statCheck : ;
    ELSE Result := Result + #13#10 + StatString(fUndo);
  END;
end;

procedure TChangeRegObj.ReportHTML(var T: TextFile;
  VAR MainS: String);
begin
  IF MainS <> fLoc+fItm THEN
    begin
      MainS := fLoc+fItm;
      WriteLn(T, '<H4>', MainS, '</H4>');
    end;
  CASE fTyp OF
    chgRegKey   : WriteLn(T, 'Changed: Key name<BR>');
    chgRegValue : WriteLn(T, 'Changed: Value name<BR>');
    chgRegData  : WriteLn(T,
      Format('Changed: Data for value "%s"<BR>', [fSub2]));
  END;
  WriteLn(T, b,b,'Old: ', fOld, '<BR>');
  WriteLn(T, b,b,'New: ', fNew, '<BR>');
  CASE fUndo AND $F OF
    statBlank : ;
    statCheck : ;
    ELSE WriteLn(T, b,b, StatString(fUndo), '<BR>');
  END;
end;

procedure TChangeRegObj.ReportText(var T: TextFile;
  VAR MainS: String);
begin
  IF MainS <> fLoc+fItm THEN
    begin
      MainS := fLoc+fItm;
      WriteLn(T);
      WriteLn(T, MainS);
    end;
  CASE fTyp OF
    chgRegKey   : WriteLn(T, #9'Changed: Key name');
    chgRegValue : WriteLn(T, #9'Changed: Value name');
    chgRegData  : WriteLn(T,
      Format(#9'Changed: Data for value "%s"', [fSub2]));
  END;
  WriteLn(T, #9#9'Old:'#9, fOld);
  WriteLn(T, #9#9'New:'#9, fNew);
  CASE fUndo AND $F OF
    statBlank : ;
    statCheck : ;
    ELSE WriteLn(T, #9#9, StatString(fUndo));
  END;
end;

function TChangeRegObj.Undo: Boolean;
VAR
  R     : TRegistry;
  DType : DWORD;
  DSize : DWORD;
  Rslt : DWORD;
  DBuff : PByteArray;
begin
  R := TRegistry.Create;
  WITH R DO
  try
    RootKey := fData1;
    fUndo   := statUndOK;
    IF R.OpenKey(fItm, False) THEN
      begin
        CASE fTyp OF
          chgRegKey : ; // won't happen
          chgRegValue : begin
            IF RegQueryValueEx(R.CurrentKey, PChar(fNew), nil,
              @DType, nil, @Dsize) = ERROR_SUCCESS THEN
            begin
              GetMem(DBuff, DSize);
              try
                IF RegQueryValueEx(R.CurrentKey, PChar(fNew), nil,
                  @DType, @DBuff[0], @Dsize) = ERROR_SUCCESS THEN
                begin
                  Rslt := RegSetValueEx(R.CurrentKey, PChar(fOld), 0,
                    DType, @DBuff[0], Dsize);
                  IF Rslt = ERROR_SUCCESS THEN
                    begin
                      Rslt := RegDeleteValue(R.CurrentKey,
                        PChar(fNew));
                      IF Rslt <> ERROR_SUCCESS THEN
                        fUndo := statUndoX OR statFaildName;
                    end;
                end;
              finally
                FreeMem(DBuff);
              end;
            end
          ELSE
            fUndo := statUndoX OR statFaildName;
          end;
          chgRegData : begin
            DSize := Length(fOld) + 1;
            DType := fData2;
            GetMem(DBuff, DSize);
            try
              StrPCopy(PChar(DBuff), fOld);
              Rslt := RegSetValueEx(R.CurrentKey, PChar(fSub2), 0,
                DType, @DBuff[0], Dsize);
              IF Rslt <> ERROR_SUCCESS THEN
                fUndo := statUndoX OR statFaildData;
            finally
              FreeMem(DBuff);
            end;
          end;
        END;
      end
    ELSE
      fUndo := statUndoX OR statFaildWrit;
    Result := fUndo = statUndOK;
  finally
    Free;
  end;
end;

{ THoldObj }

constructor THoldObj.Create(vCode: Integer; vType : DWORD;
  vData: String);
begin
  Inherited Create;
  fCode  := vCode;
  fType  := vType;
  fData1 := vData;
end;

end.
